# pylint: disable = import-error
import bpy
import numpy as np
import mathutils
import math
from .constants import BREAKABLE_CONSTRAINTS, UNBREAKABLE_CONSTRAINTS, \
    BREAKABLE_COLOR, UNBREAKABLE_COLOR

def findCloseObjects(obj, selectedObjects):
    if(obj.type == 'MESH'):
        
        hitObjects = []

        for poly in obj.data.polygons:  
            # get objects world space transforms and convert
            # objects polys to worldspace
            worldCoords = obj.matrix_world.decompose()
            objRot = worldCoords[1]
            direction = objRot @ poly.normal
            origin = obj.matrix_world @ poly.center + direction *0.001
            distance = 1
                    
            # test to see if any polys hit another object    
            test = bpy.context.scene.ray_cast(
                    bpy.context.view_layer.depsgraph,
                    # move ray origin away from own poly to avoid hitting its self
                    origin = origin, 
                    direction = direction,
                    distance=distance
                    )
            
            # check to see if hit object was included in original selection
            # and add to list if it is
            for x in selectedObjects:
                hitObject = test[4]
                if(x == hitObject):
                    if(test[0] and hitObjects.count(hitObject) == 0):
                        hitObjects.append(hitObject)
        
        # return the found objects
        return hitObjects
    return []

def deleteObject(obj):
    bpy.data.objects.remove (
    object = obj,
    do_unlink = True,
    do_id_user = True,
    do_ui_user = True
    )
    return 'FINISHED'


def constraintExists(object1, object2, collection):
    try:
        col = bpy.data.collections[collection]
        for ob in col.objects:
                existingObs = [ob.rigid_body_constraint.object1, ob.rigid_body_constraint.object2]
                if(existingObs.count(object1) > 0 and existingObs.count(object2) > 0):
                    return True
    except:
        return False

def createCollection(colName):
    newCol = bpy.data.collections.new(colName)
    return newCol


## functions end-------------------------------------------------------------- ##

## ------------------------------ main ----------------------------------------##

def main(context, props):

    wm_progress = context.window_manager
    wm_progress.progress_begin (
        min = 0,
        max = 100
    )

    breakableCol = BREAKABLE_CONSTRAINTS in bpy.data.collections
    unbreakableCol = UNBREAKABLE_CONSTRAINTS in bpy.data.collections
    masterCollection = context.scene.collection
    # set up some new collections
    if(not breakableCol):
        newCol = createCollection(BREAKABLE_CONSTRAINTS)
        masterCollection.children.link(newCol)
    
    if(not unbreakableCol):
        newCol = createCollection(UNBREAKABLE_CONSTRAINTS)
        masterCollection.children.link(newCol)

    # set the current active collection to master
    context.view_layer.active_layer_collection = context.view_layer.layer_collection

    makeBreakable = props.MakeBreakable
    glueStrength = props.GlueStrength
    glueStrengthRandom = props.GlueStrengthRandom
    varMin = props.VarMin
    varMax = props.VarMax
    maxConstraints = props.MaxConstraints
    method = props.ConstraintMethod

    selectedObjects = context.selected_objects
    
    for index, obj in enumerate(selectedObjects):

        methodTypes = {
            "STD": None,
            "DIST": None,
            "PROX": findCloseObjects 
        }

        func = methodTypes[method]
        
        conObjs = func(obj, selectedObjects)
        
        # if user wants to limit number of connections
        if(maxConstraints < 100):
            total = len(conObjs)
            reduction = round(total * (maxConstraints/100))
            for i in range(total - reduction):
                item = np.random.choice(conObjs, replace = False)
                conObjs.remove(item)

        if(glueStrengthRandom):
            random = np.random.uniform(varMin, varMax) 
            glueStrength = random
       
        # make new contraints
        for conObj in conObjs:
            try:
                isRigidBody1 = obj.rigid_body
                isRigidBody2 = conObj.rigid_body
                if(isRigidBody1 and isRigidBody2):
                    wm_progress.progress_update(index/len(selectedObjects))
                    # check to see if an existing contraint exists between objects during creation...
                    existingBreakable = constraintExists(obj, conObj, BREAKABLE_CONSTRAINTS)
                    existingUnbreakable = constraintExists(obj, conObj, UNBREAKABLE_CONSTRAINTS)
                    # ..and skip creation if true 
                    if(not existingBreakable and not existingUnbreakable):
                        bpy.ops.object.empty_add(
                            type='PLAIN_AXES', 
                            radius=1.0, align='WORLD', 
                            location=obj.location, 
                            rotation=(0.0, 0.0, 0.0))
                        
                        newEmpty = context.view_layer.objects.active

                        bpy.ops.rigidbody.constraint_add()

                        newEmpty.name = 'pointConstraint'

                        masterCollection.objects.unlink(newEmpty)

                        if(makeBreakable):
                            bpy.data.collections[BREAKABLE_CONSTRAINTS].objects.link(newEmpty)
                            newEmpty.rigid_body_constraint.use_breaking = makeBreakable
                            newEmpty.rigid_body_constraint.breaking_threshold = glueStrength
                            obj.color = BREAKABLE_COLOR
                            conObj.color = BREAKABLE_COLOR
                        else:
                            bpy.data.collections[UNBREAKABLE_CONSTRAINTS].objects.link(newEmpty)
                            obj.color = UNBREAKABLE_COLOR
                            conObj.color = UNBREAKABLE_COLOR

                        newEmpty.rigid_body_constraint.object1 = obj

                        newEmpty.rigid_body_constraint.object2 = conObj 
            except:
                print("can't create constraint objects")
    context.window_manager.progress_end()